/* ********************************************************************** */
/* example1.c: using APRON library  */
/* ********************************************************************** */

/* We illustrate here the use of the APRON interface, at level 1 */

#include "ap_global0.h"
#include "ap_global1.h"

#include "box.h"
#include "oct.h"
#include "pk.h"
#include "pkeq.h"
#include "ap_ppl.h"

void ex1(ap_manager_t* man)
{
  printf("******************************\n");
  printf("ex1\n");
  printf("Library %s, version %s\n",man->library,man->version);
  printf("******************************\n");

  ap_var_t name_of_dim[6] = {
    "x","y","z", "u", "v", "w"
  };
  ap_environment_t* env = ap_environment_alloc(&name_of_dim[0],3,&name_of_dim[3],3);

  /* =================================================================== */
  /* Creation of polyhedra
     x=1, y=1, z=2 */
  /* =================================================================== */

  /* 0. Create the array */
  ap_lincons1_array_t array = ap_lincons1_array_make(env,3);

  /* 1.a Creation of an equality constraint x=1 */
  ap_linexpr1_t expr = ap_linexpr1_make(env,AP_LINEXPR_SPARSE,0);
  ap_lincons1_t cons = ap_lincons1_make(AP_CONS_EQ,&expr,NULL);
    /* Now expr is memory-managed by cons */

  /* 1.b Fill the constraint */
  ap_lincons1_set_list(&cons,AP_COEFF_S_INT,1,"x",AP_CST_S_INT,-1,AP_END);
  /* 1.c Put in the array */
  ap_lincons1_array_set(&array,0,&cons);
    /* Now cons is memory-managed by array */

  /* 1.a Creation of an equality constraint y=1 */
  expr = ap_linexpr1_make(env,AP_LINEXPR_SPARSE,0);
  cons = ap_lincons1_make(AP_CONS_EQ,&expr,NULL);
  ap_lincons1_set_list(&cons,AP_COEFF_S_INT,1,"y",AP_CST_S_INT,-1,AP_END);
  ap_lincons1_array_set(&array,1,&cons);

  /* 1.a Creation of an equality constraint z=2 */
  expr = ap_linexpr1_make(env,AP_LINEXPR_SPARSE,0);
  cons = ap_lincons1_make(AP_CONS_EQ,&expr,NULL);
  ap_lincons1_set_list(&cons,AP_COEFF_S_INT,1,"z",AP_CST_S_INT,-2,AP_END);
  ap_lincons1_array_set(&array,2,&cons);

  /* 2. Creation of an abstract value */
  ap_abstract1_t abs1 = ap_abstract1_of_lincons_array(man,env,&array);

  fprintf(stdout,"Abstract value 1:\n");
  ap_abstract1_fprint(stdout,man,&abs1);

  /* deallocation */
  ap_lincons1_array_clear(&array);

  /* =================================================================== */
  /* Creation of polyhedra
     x=1, y=2, z=3 */
  /* =================================================================== */
  /* 1.a Creation of an equality constraint x=1 */
  array = ap_lincons1_array_make(env,3);
  expr = ap_linexpr1_make(env,AP_LINEXPR_SPARSE,0);
  cons = ap_lincons1_make(AP_CONS_EQ,&expr,NULL);
  ap_linexpr1_set_list(&expr,AP_COEFF_S_INT,1,"x",AP_CST_S_INT,-1,AP_END);
  ap_lincons1_array_set(&array,0,&cons);

  /* 1.a Creation of an equality constraint y=2 */
  expr = ap_linexpr1_make(env,AP_LINEXPR_SPARSE,0);
  cons = ap_lincons1_make(AP_CONS_EQ,&expr,NULL);
  ap_linexpr1_set_list(&expr, AP_COEFF_S_INT,1,"y",AP_CST_S_INT,-2,AP_END);
  ap_lincons1_array_set(&array,1,&cons);

  /* 1.a Creation of an equality constraint z=3 */
  expr = ap_linexpr1_make(env,AP_LINEXPR_SPARSE,0);
  cons = ap_lincons1_make(AP_CONS_EQ,&expr,NULL);
  ap_linexpr1_set_list(&expr, AP_COEFF_S_INT,1,"z",AP_CST_S_INT,-3, AP_END);
  ap_lincons1_array_set(&array,2,&cons);

  /* 2. Creation of an abstract value */
  ap_abstract1_t abs2 = ap_abstract1_of_lincons_array(man,env,&array);

  fprintf(stdout,"Abstract value 2:\n");
  ap_abstract1_fprint(stdout,man,&abs2);

  ap_lincons1_array_clear(&array);

  /* =================================================================== */
  /* Creation of polyhedra
     x=2, y=3, z=5 */
  /* =================================================================== */
  /* 1.a Creation of an equality constraint x=2 */
  array = ap_lincons1_array_make(env,3);
  expr = ap_linexpr1_make(env,AP_LINEXPR_SPARSE,0);
  cons = ap_lincons1_make(AP_CONS_EQ,&expr,NULL);
  ap_linexpr1_set_list(&expr,AP_COEFF_S_INT,1,"x",AP_CST_S_INT,-2,AP_END);
  ap_lincons1_array_set(&array,0,&cons);

  /* 1.a Creation of an equality constraint y=3 */
  expr = ap_linexpr1_make(env,AP_LINEXPR_SPARSE,0);
  cons = ap_lincons1_make(AP_CONS_EQ,&expr,NULL);
  ap_linexpr1_set_list(&expr, AP_COEFF_S_INT,1,"y",AP_CST_S_INT,-3,AP_END);
  ap_lincons1_array_set(&array,1,&cons);

  /* 1.a Creation of an equality constraint z=5 */
  expr = ap_linexpr1_make(env,AP_LINEXPR_SPARSE,0);
  cons = ap_lincons1_make(AP_CONS_EQ,&expr,NULL);
  ap_linexpr1_set_list(&expr, AP_COEFF_S_INT,1,"z",AP_CST_S_INT,-5, AP_END);
  ap_lincons1_array_set(&array,2,&cons);

  /* 2. Creation of an abstract value */
  ap_abstract1_t abs3 = ap_abstract1_of_lincons_array(man,env,&array);

  fprintf(stdout,"Abstract value 3:\n");
  ap_abstract1_fprint(stdout,man,&abs3);

  ap_lincons1_array_clear(&array);

  /* =================================================================== */
  /* Creation of polyhedra
     x=3, y=5, z=8 */
  /* =================================================================== */
  /* 1.a Creation of an equality constraint x=3 */
  array = ap_lincons1_array_make(env,3);
  expr = ap_linexpr1_make(env,AP_LINEXPR_SPARSE,0);
  cons = ap_lincons1_make(AP_CONS_EQ,&expr,NULL);
  ap_linexpr1_set_list(&expr,AP_COEFF_S_INT,1,"x",AP_CST_S_INT,-3,AP_END);
  ap_lincons1_array_set(&array,0,&cons);

  /* 1.a Creation of an equality constraint y=5 */
  expr = ap_linexpr1_make(env,AP_LINEXPR_SPARSE,0);
  cons = ap_lincons1_make(AP_CONS_EQ,&expr,NULL);
  ap_linexpr1_set_list(&expr, AP_COEFF_S_INT,1,"y",AP_CST_S_INT,-5,AP_END);
  ap_lincons1_array_set(&array,1,&cons);

  /* 1.a Creation of an equality constraint z=8 */
  expr = ap_linexpr1_make(env,AP_LINEXPR_SPARSE,0);
  cons = ap_lincons1_make(AP_CONS_EQ,&expr,NULL);
  ap_linexpr1_set_list(&expr, AP_COEFF_S_INT,1,"z",AP_CST_S_INT,-8, AP_END);
  ap_lincons1_array_set(&array,2,&cons);

  /* 2. Creation of an abstract value */
  ap_abstract1_t abs4 = ap_abstract1_of_lincons_array(man,env,&array);

  fprintf(stdout,"Abstract value 4:\n");
  ap_abstract1_fprint(stdout,man,&abs4);

  ap_lincons1_array_clear(&array);


  /* =================================================================== */
  /* Join and Widening */
  /* =================================================================== */

  ap_abstract1_t res = ap_abstract1_join(man,false,&abs2,&abs3);
  fprintf(stdout,"Join (2U3):\n");
  ap_abstract1_fprint(stdout,man,&res);

  ap_abstract1_t res2 = ap_abstract1_join(man,false,&res,&abs1);
  fprintf(stdout,"Join (1U2U3):\n");
  ap_abstract1_fprint(stdout,man,&res2);

  ap_abstract1_t res3 = ap_abstract1_join(man,false,&res,&abs4);
  fprintf(stdout,"Join (2U3U4):\n");
  ap_abstract1_fprint(stdout,man,&res3);
  
  ap_abstract1_t res4 = ap_abstract1_widening(man,&res2,&res3);
  fprintf(stdout,"Widening (1U2U3) and (2U3U4):\n");
  ap_abstract1_fprint(stdout,man,&res4);
  ap_abstract1_clear(man,&res4);
  ap_abstract1_clear(man,&res);

  res4 = ap_abstract1_join(man,false,&res2,&abs4);
  fprintf(stdout,"Join (1U(2U3))U4:\n");
  ap_abstract1_fprint(stdout,man,&res4);
  ap_abstract1_clear(man,&res4);
  ap_abstract1_clear(man,&res2);
  ap_abstract1_clear(man,&res3);

  res = ap_abstract1_join(man,false,&abs1,&abs2);
  fprintf(stdout,"Join (1U2):\n");
  ap_abstract1_fprint(stdout,man,&res);

  res2 = ap_abstract1_join(man,false,&abs3,&abs4);
  fprintf(stdout,"Join (3U4):\n");
  ap_abstract1_fprint(stdout,man,&res2);

  res3 = ap_abstract1_join(man,false,&res,&res2);
  fprintf(stdout,"Join (1U2 U 3U4):\n");
  ap_abstract1_fprint(stdout,man,&res3);
  ap_abstract1_clear(man,&res);
  ap_abstract1_clear(man,&res2);
  ap_abstract1_clear(man,&res3);


  /* deallocation */
  ap_abstract1_clear(man,&abs1);
  ap_abstract1_clear(man,&abs2);
  ap_abstract1_clear(man,&abs3);
  ap_abstract1_clear(man,&abs4);

  ap_environment_free(env);
}





int main(int argc, const char** argv)
{
  ap_manager_t* man;

  if (argc > 1)
  	printf("%s\n",argv[0]);

  man = box_manager_alloc();
  ex1(man);
  ap_manager_free(man);

  man = oct_manager_alloc();
  ex1(man);
  ap_manager_free(man);

  man = pk_manager_alloc(true);
  ex1(man);
  ap_manager_free(man);

  man = ap_ppl_poly_manager_alloc(false);
  ex1(man);
  ap_manager_free(man);

  man = pkeq_manager_alloc();
  ex1(man);
  ap_manager_free(man);

  man = ap_ppl_grid_manager_alloc();
  ex1(man);
  ap_manager_free(man);

}
